###############################################################################
#
#    NAME: wkcfdem.tcl
#
#    PURPOSE: Useful procedures to write the properties of DEM application to the Kratos *.mdpa
#
#    QUANTECH ATZ-DEVELOPMENT DEPARTMENT
#
#    AUTHOR : G. Socorro
#
#    CREATED AT: 01/10/13
#
#    HISTORY:
#
#     0.6- 31/01/14-G. Socorro, correct a bug in the proc WriteDEMConditionProperties (write the value of IMPOSED_VELOCITY_X_VALUE, etc.)
#     0.5- 23/01/14-J. Garate, procs GetDEMGroupMeshProperties, ::wkcf::WriteDEMGroupMeshProperties, WriteDampingData,WriteMatTestData
#                              , WritePostProcessData, WriteMaterialModelData,WriteExplicitSolverVariables created or modified
#     0.4- 08/11/13-G. Socorro, modify some variable identifier (add DEM-*)
#     0.3- 03/11/13-G. Socorro, add the proc GetInletGroupNodes to get the inlet condition group node list
#     0.2- 02/11/13-G. Socorro, add the proc AssignSpecialBoundaries and GetBoundariesNodeList
#     0.1- 01/10/13-G. Socorro, create a base source code from wkcf.tcl
#
###############################################################################

proc ::wkcf::GetBoundariesNodeList {} {
    variable ndime

    set nlist [list]
    set cgroupid "-AKGDEMSkinMesh2D"
    if {$ndime =="3D"} {
	set cgroupid "-AKGDEMSkinMesh3D"
    }
    if { [GiD_Groups exists $cgroupid] } {
	if {[GiD_EntitiesGroups get $cgroupid nodes -count]} {
	    set nlist [GiD_EntitiesGroups get $cgroupid nodes]
	    #wa $nlist
	}
    }
    return $nlist
}

proc ::wkcf::GetDEMActiveElements { } {
    variable ActiveAppList
    variable ndime
    
    set gelemlist ""
    set basexpath "DEM//c.DEM-Elements"
    set elembasexpath "$basexpath//c.DEM-Element"
    
    if {"Fluid" in $ActiveAppList} {
	set gelemlist [list "SwimmingDEMElement"]
    } else {
	if {$ndime eq "3D"} {
		set grouplist [::xmlutils::setXmlContainerIds $elembasexpath]
		if {$grouplist != ""} {
		    set gelemlist [list "SphericPartDEMElement3D"]
		set cxpath "DEM//c.DEM-Cohesivegroup"
		set grouplist [::xmlutils::setXmlContainerIds $cxpath]
		foreach cgroup $grouplist {
		    set cproperty "dv"
		    set cxpath "DEM//c.DEM-Cohesivegroup//c.[list ${cgroup}]//c.Properties//i.CohGroup"
		    set cohesive_group_number [::xmlutils::setXml $cxpath $cproperty]
		    if {$cohesive_group_number != 0} {
		    set gelemlist [list "SphericContPartDEMElement3D"]
		       
		       break
		    }
		}
	    }
	} else {
		set grouplist [::xmlutils::setXmlContainerIds $elembasexpath]
		if {$grouplist != ""} {
		set gelemlist [list "CylPartDEMElement2D"]
		set cxpath "DEM//c.DEM-Cohesivegroup"
		set grouplist [::xmlutils::setXmlContainerIds $cxpath]
		foreach cgroup $grouplist {
		    set cproperty "dv"
		    set cxpath "DEM//c.DEM-Cohesivegroup//c.[list ${cgroup}]//c.Properties//i.CohGroup"
		    set cohesive_group_number [::xmlutils::setXml $cxpath $cproperty]
		    if {$cohesive_group_number != 0} {
		       set gelemlist [list "CylContPartDEMElement2D"]
		       break
		    }
		}
	    }
	}
    }
    return $gelemlist
}

proc ::wkcf::AssignSpecialBoundaries {ndime entitylist} {

    # wa "ndime:$ndime entitylist:$entitylist"
    set DEMApplication "No"
    set cproperty "dv"
    set cxpath "GeneralApplicationData//c.ApplicationTypes//i.DEM"
    catch { set DEMApplication [::xmlutils::setXml $cxpath $cproperty] }
    if {$DEMApplication eq "Yes"} {
	if { $ndime =="2D" } {        
	    # Automatic Kratos Group for all DEM boundary points
	    set groupid "-AKGDEMSkinMesh2D"
	    set entitytype "point" 
	    ::wkcf::CleanAutomaticConditionGroupGiD $entitytype $groupid        
	    # Get all end point list from the boundary lines
	    set endpointlist [list]
	    foreach lineid $entitylist {
		lappend endpointlist {*}[lrange [GiD_Geometry get line $lineid] 2 end]
	    }
	    set endpointlist [lsort -integer -unique $endpointlist]
	    # wa "endpointlist:$endpointlist"
	    # Assign the boundary condition
	    ::wkcf::AssignConditionToGroupGID $entitytype $endpointlist $groupid 

	} elseif { $ndime =="3D" } {

	    # Automatic Kratos Group for all DEM boundary lines
	    set groupid "-AKGDEMSkinMesh3D"
	    set entitytype "line" 
	    ::wkcf::CleanAutomaticConditionGroupGiD $entitytype $groupid
	    # Get all end line list from the boundary surfaces
	    set endlinelist [list]
	    foreach surfid $entitylist {
		set surfprop [GiD_Geometry get surface $surfid]
		# set surfacetype [lindex $surfprop 0]
		set nline [lindex $surfprop 2]
		# wa "surfacetype:$surfacetype\nnline:$nline\nsurfprop:$surfprop"
		set lineprop [list]
		#if {$surfacetype eq "nurbssurface"} {
		    set lineprop [lrange $surfprop 9 [expr {9+$nline-1}]]
		#}
		foreach lprop $lineprop {
		    lassign $lprop lineid orientation 
		    lappend endlinelist $lineid
		}
	    }
	    # wa "before endlinelist:$endlinelist"
	    set endlinelist [lsort -integer -unique $endlinelist]
	    # wa "endlinelist:$endlinelist"
	    # Assign the boundary condition
	    ::wkcf::AssignConditionToGroupGID $entitytype $endlinelist $groupid
	}
    }
}

proc ::wkcf::GetInletGroupMeshProperties {AppId} {
    # ABSTRACT: Get the inlet condition group properties mdpa file (only the nodes)
    variable dprops
   
    # For debug
    if {!$::wkcf::pflag} {
        set inittime [clock seconds]
    }

    # Write assigned group to the elements
    set meshgroupid 0

    # Init the global mesh identifier list
    set dprops($AppId,AllMeshId) [list]
    set dprops($AppId,AllMeshGroupId) [list]

    # Set the rootid
    set rootid "$AppId"
    set cproperty "dv"
    
    # Get the values
    set basexpath "$rootid//c.DEM-Conditions//c.DEM-Inlet"
    set gproplist [::xmlutils::setXmlContainerIds $basexpath]
    # wa "gproplist:$gproplist"
    foreach cgroupid $gproplist {
	# Get the group properties
	set cxpath "${basexpath}//c.[list ${cgroupid}]//c.MainProperties"
	set allgprop [::xmlutils::setXmlContainerPairs $cxpath "" "dv"]
	# wa "allgprop:$allgprop"
	if {[llength $allgprop]} {
	    if {[GiD_EntitiesGroups get $cgroupid elements -count]} {
		incr meshgroupid 1
		# Create the meshid-group identifier mapping
		lappend dprops($AppId,AllMeshId) $meshgroupid
		# Set mesh-group id mapping
		if {$cgroupid ni $dprops($AppId,AllMeshGroupId)} {
		    lappend dprops($AppId,AllMeshGroupId) $cgroupid
		}
		set dprops($AppId,Mesh,$cgroupid,MeshIdGroup) $meshgroupid
		
		# Set the properties
		# Modify some properties
		set newprops [list]
		foreach propid $allgprop {
		    lassign $propid key value
		    if {($key eq "MatModel")||($key eq "DirectionVectorX") ||($key eq "DirectionVectorY") ||($key eq "DirectionVectorZ")} {
		        continue
		    } else {
		        if {$key eq "VelocityModulus"} {
		            set uservelmod $value
		            # Find velocity module 
		            set findvelmol [lsearch -index 0 $allgprop "VelocityModulus"]
		            # wa "findvelmol:$findvelmol"
		            set findvx [lsearch -index 0 $allgprop "DirectionVectorX"]
		            # wa "findvx:$findvx"
		            set findvy [lsearch -index 0 $allgprop "DirectionVectorY"]
		            # wa "findvy:$findvy"
		            set findvz [lsearch -index 0 $allgprop "DirectionVectorZ"]
		            # wa "findvz:$findvz"
		            if {($findvx !="-1") && ($findvy !="-1")&& ($findvz !="-1")} {
		                # Get the velocity values
		                set vxval [lindex $allgprop $findvx 1]
		                set vyval [lindex $allgprop $findvy 1]
		                set vzval [lindex $allgprop $findvz 1]
		                set velvector [list $vxval $vyval $vzval] 
		                # Calculate the modulus
		                set velmod [::MathUtils::VectorModulus $velvector]
		                # wa "velmod:$velmod"
		                set endvelmodvalue "\[3\] (0.0,0.0,0.0)"
		                if {$velmod !=0.0} {
		                    set vxval [expr {($vxval*$uservelmod)/$velmod}]
		                    set vyval [expr {($vyval*$uservelmod)/$velmod}]
		                    set vzval [expr {($vzval*$uservelmod)/$velmod}]
		                    # wa "vxval:$vxval vyval:$vyval vzval:$vzval"
		                    set endvelmodvalue "\[3\] ($vxval,$vyval,$vzval)"
		                } 
		                lappend newprops [list $key $endvelmodvalue]        
		            }
		        } elseif {$key eq "ParticleDiameter"} {
		            set finddiameter [lsearch -index 0 $allgprop "ParticleDiameter"]
		            if {$finddiameter !="-1"} {
                        set diametervalue [lindex $allgprop $finddiameter 1]
                        set radius [expr $diametervalue * 0.5]
                        lappend newprops [list "ParticleRadius" $radius]  
		            } else {
		                wa "No diameter found in DEM inlet $cgroupid"
		            }
		        } else {
		            lappend newprops [list $key $value]        
		        }
		    }
		}
		# wa "newprops:$newprops"
		if {[llength $newprops]} {
		    set dprops($AppId,Mesh,$cgroupid,MeshIdGroupProp) $newprops
		}
	    }
	}
    }
    
    # For debug
    if {!$::wkcf::pflag} {
	set endtime [clock seconds]
	set ttime [expr $endtime-$inittime]
	# WarnWinText "endtime:$endtime ttime:$ttime"
	WarnWinText "Obtening DEM-Inlet group using the new mesh format: [::KUtils::Duration $ttime]"
    }   
}

proc ::wkcf::GetInletGroupNodes {AppId cgroupid} {
    # ABSTRACT: Get the inlet condition group node list
    variable dprops
    
    set cprop [list]

    # For debug
    if {!$::wkcf::pflag} {
        set inittime [clock seconds]
    }
    
    # Set the rootid
    set rootid "$AppId"
    set cproperty "dv"
    
    # Get the values
    set basexpath "$rootid//c.DEM-Conditions//c.DEM-Inlet"
    # Get the group properties
    set cxpath "${basexpath}//c.[list ${cgroupid}]//c.MainProperties"
    set allgprop [::xmlutils::setXmlContainerPairs $cxpath "" "dv"]
    # wa "allgprop:$allgprop"
    if {[llength $allgprop]} {
	if {[GiD_EntitiesGroups get $cgroupid nodes -count]} {
	    # Check for the exclude boundaries option
	    set findeb [lsearch -index 0 $allgprop "ExcludeBoundaries"]
	    # wa "findeb:$findeb"
	    set ExcludeBoundaries "No"
	    if {$findeb !="-1"} {
		set ExcludeBoundaries [lindex $allgprop $findeb 1]
	    }
	    # wa "ExcludeBoundaries:$ExcludeBoundaries"
	    if {$ExcludeBoundaries eq "No"} {
		set cprop [GiD_EntitiesGroups get $cgroupid nodes]
	    } else {
		# Get the boundary node list 
		set nlist [::wkcf::GetBoundariesNodeList]
		foreach node_id [GiD_EntitiesGroups get $cgroupid nodes] {
		    set findnode [lsearch $nlist $node_id]
		    # wa "findnode:$findnode"
		    if {$findnode =="-1"} {
		        lappend cprop $node_id
		    }
		}
	    }
	}
    }
    
    # For debug
    if {!$::wkcf::pflag} {
	set endtime [clock seconds]
	set ttime [expr {$endtime-$inittime}]
	WarnWinText "Get DEM-Inlet group nodes list: [::KUtils::Duration $ttime]"
    }
    return $cprop  
}


proc ::wkcf::GetDemFemWallGroupNodes {AppId cgroupid} {
    # ABSTRACT: Get the inlet condition group node list
    variable dprops
    
    set cprop [list]

    # For debug
    if {!$::wkcf::pflag} {
        set inittime [clock seconds]
    }
    
    # Set the rootid
    set rootid "$AppId"
    set cproperty "dv"
    
    # Get the values
    set basexpath "$rootid//c.DEM-Conditions//c.DEM-FEM-Wall"
    # Get the group properties
    #set cxpath "${basexpath}//c.[list ${cgroupid}]//c.MainProperties"
    #set allgprop [::xmlutils::setXmlContainerPairs $cxpath "" "dv"]
    # wa "allgprop:$allgprop"
    #if {[llength $allgprop]} {

      if {[GiD_EntitiesGroups get $cgroupid elements -count]} {        
        set elems [GiD_EntitiesGroups get $cgroupid elements]
	foreach eid $elems {
	    # JG cambiar a GiD_Mesh
		set nodes [lrange [GiD_Mesh get element $eid] 3 end]
		# wa "nodes on demfem $nodes"
		# JG OJO REPETIDOS LSORT + FLAGS
		    lappend cprop {*}$nodes
	    
	}
      }
    #}
    
    # For debug
    if {!$::wkcf::pflag} {
    set endtime [clock seconds]
    set ttime [expr {$endtime-$inittime}]
    WarnWinText "Get DEM-FEM-Wall group nodes list: [::KUtils::Duration $ttime]"
    }
    return [lsort -increasing -integer -unique $cprop]
}



proc ::wkcf::WriteDEMGroupMeshProperties {AppId} {
    # ABSTRACT: Write all DEM condition group properties mdpa file (only the nodes)
    variable filechannel
    # wa "WriteDEMGroupMeshProperties $AppId"
    global KPriv
    # For debug
    if {!$::wkcf::pflag} {
        set inittime [clock seconds]
    }
    
    set basexpath "$AppId//c.DEM-Conditions//c.DEM-FEM-Wall"
    set gFEMproplist [::xmlutils::setXmlContainerIds $basexpath]
    
    set cproperty "dv"
    set meshgroupid 0
    
    set condlist [list "DEM-ForceIntegrationGroup" "DEM-TopLayerGroup" \
		        "DEM-BottomLayerGroup" "DEM-VelocityBC"]
    
    set pathlist [list "$AppId//c.DEM-Results//c.DEM-Graphs//c.DEM-ForceIntegrationGroup" \
		        "$AppId//c.DEM-MaterialTest//c.DEM-TopLayerGroup" \
		        "$AppId//c.DEM-MaterialTest//c.DEM-BottomLayerGroup" \
		        "$AppId//c.DEM-Conditions//c.DEM-VelocityBC"]
    

    
    foreach condid $condlist cpath $pathlist {
	set gproplist [::xmlutils::setXmlContainerIds $cpath]
	foreach cgroupid $gproplist {
	    # wa "condid: $condid cgroupid: $cgroupid"
	    if {$cgroupid ni $gFEMproplist} {
	    if {[GiD_EntitiesGroups get $cgroupid elements -count]} {
		    if { [lindex [GiD_Mesh get element [lindex [GiD_EntitiesGroups get $cgroupid elements] 0] ] 1] eq "Sphere"} {
			# wa "condid: $condid cgroupid: $cgroupid"
		    incr meshgroupid 1
			# Ai Vi FIG Top Bottom
			set valuelist [list 0 0.0 0 0.0 0 0.0 0 0 0]
		# Write mesh properties for this group
			GiD_File fprintf $filechannel "%s" "Begin Mesh $meshgroupid \/\/ GUI conditionid $condid group identifier: $cgroupid"
		
		if {$condid eq "DEM-VelocityBC"} {
		    set localpath  "$cpath//c.$cgroupid"
		    set allgprop [::xmlutils::setXmlContainerPairs $localpath "" "dv"]
		    # wa "allgprop:$allgprop"
		    set Ax 0; set Ay 0; set Az 0; set Vx 0.0; set Vy 0.0; set Vz 0.0
		    foreach var $allgprop {
		        lassign $var cvar cval
		        set $cvar $cval
		    }
		    lset valuelist 0 $Ax
		    lset valuelist 1 $Vx
		    lset valuelist 2 $Ay
		    lset valuelist 3 $Vy
		    lset valuelist 4 $Az
		    lset valuelist 5 $Vz
		    # wa "valuelist:$valuelist"
		} else {
			    if {$condid eq "DEM-ForceIntegrationGroup"} {
		    lset valuelist 6 1
		}
			    if {$condid eq "DEM-TopLayerGroup"} {
				lset valuelist 6 1
				lset valuelist 7 1
			    }
			    if {$condid eq "DEM-BottomLayerGroup"} {
				lset valuelist 6 1
				lset valuelist 8 1
			    }
			}
			# wa "Condid $condid valuelist $valuelist"
		# Write Properties
		::wkcf::WriteDEMConditionProperties $AppId $cgroupid $valuelist
		
		# Write nodes
		GiD_File fprintf $filechannel "%s" " "
		GiD_File fprintf $filechannel "%s" " Begin MeshNodes"
		foreach eid [GiD_EntitiesGroups get $cgroupid elements] {
		    # JG cambiar a Mesh
		    set node_id [lindex [GiD_Info Mesh Elements sphere $eid $eid] 1]
	    
		    GiD_File fprintf $filechannel "%10i" $node_id
		}
        GiD_File fprintf $filechannel "%s" " End MeshNodes"
        GiD_File fprintf $filechannel "%s" " "
        GiD_File fprintf $filechannel "%s" "End Mesh"
        GiD_File fprintf $filechannel "%s" ""
	    }
	}
    }
	}
    }
   
    # Write Mesh Properties for DEM FEM Wall Condition
    ::wkcf::WriteDEMFEMWallMeshProperties $AppId
    
    # For debug
    if {!$::wkcf::pflag} {
	set endtime [clock seconds]
	set ttime [expr {$endtime-$inittime}]
	# WarnWinText "endtime:$endtime ttime:$ttime"
	WarnWinText "Write DEM-Mesh Properties group: [::KUtils::Duration $ttime]"
    }   
}


proc ::wkcf::WriteBoundingBoxDefaults {fileid} {

    puts $fileid "BoundingBoxMaxX                  = 3.00000e+00"  
    puts $fileid "BoundingBoxMaxY                  = 3.00000e+00"
    puts $fileid "BoundingBoxMaxZ                  = 2.00000e+01"
    puts $fileid "BoundingBoxMinX                  = -3.00000e+00"
    puts $fileid "BoundingBoxMinY                  = -3.00000e+00"
    puts $fileid "BoundingBoxMinZ                  = 0.00000e+00"

}

proc ::wkcf::WriteDampingData {fileid} {
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-Damping//i.DEM-LocalContactDamping"
    set LocalContact [::xmlutils::setXml $cxpath "dv"]
    
    if {$LocalContact eq "None"} {
	puts $fileid "LocalContactDamping                  = \"None\""
    } else {
	if {$LocalContact eq "Normal"} {
	    puts $fileid "LocalContactDamping              = \"Normal\""
	} elseif {$LocalContact eq "Tangential"} {
	    puts $fileid "LocalContactDamping              = \"Tangential\""
	} elseif {$LocalContact eq "Both"} {
	    puts $fileid "LocalContactDamping              = \"Both\""
	}
	set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-Damping//i.DEM-LocalDampingFactor"
	set LocalDampingFactor [::xmlutils::setXml $cxpath "dv"]
	puts $fileid "LocalDampingFactor               = $LocalDampingFactor"
    }
    
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-Damping//i.DEM-GlobalDynamicsDamping"
    set GlobalDump [::xmlutils::setXml $cxpath "dv"]
    
    if {$GlobalDump eq "No"} {
	puts $fileid "GlobalForceReduction             = 0.0"
    } elseif {$GlobalDump eq "Yes"} {
	set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-Damping//i.DEM-GlobalForceReduction"
	set GlobForceRed [::xmlutils::setXml $cxpath "dv"]
	puts $fileid "GlobalForceReduction                  = $GlobForceRed"
    }
}

proc ::wkcf::WriteMatTestData {fileid} {
    
    set cxpath "DEM//c.DEM-Options//c.DEM-AdvancedOptions//i.DEM-MaterialTestActivate"
    set TestTypeOn [::xmlutils::setXml $cxpath "dv"]

	set cxpath "DEM//c.DEM-MaterialTest//i.DEM-TestType"
	set TestType [::xmlutils::setXml $cxpath "dv"]

    if {$TestTypeOn eq "Show"} {
	puts $fileid "TestType                              = \"$TestType\""
    } else {
    puts $fileid "TestType                              = \"None\""
    }
	
    set LVel 0.0
	set cxpath "DEM//c.DEM-MaterialTest//i.DEM-ConfinementPressure"
	set ConfPress [::xmlutils::setXml $cxpath "dv"]
	puts $fileid "ConfinementPressure                   = $ConfPress"
	
    set basexpath "DEM//c.DEM-MaterialTest//c.DEM-TopLayerGroup"
    set topgroup [::xmlutils::setXmlContainerIds $basexpath]
    if {[llength $topgroup]} {
	set basexpath "DEM//c.DEM-Conditions//c.DEM-FEM-Wall"
	set gproplist [::xmlutils::setXmlContainerIds $basexpath]
	if {[llength $gproplist]} {
	    set cxpath "DEM//c.DEM-Conditions//c.DEM-FEM-Wall//c.[lindex $topgroup 0]//c.LinearVelocity//i.LinearVelocityY"
	    #wa "TopGroup path $cxpath val $dv"
	set LVel [::xmlutils::setXml $cxpath "dv"]
	}
    }
	puts $fileid "LoadingVelocityTop                             = $LVel"
    
	puts $fileid "LoadingVelocityBot                             = 0.0"
    
    puts $fileid "FemPlates                             = \"OFF\""
  puts $fileid "StressStrainOption                    = \"OFF\""

	set cxpath "DEM//c.DEM-MaterialTest//i.DEM-MeshType"
	set mt [::xmlutils::setXml $cxpath "dv"]
	puts $fileid "MeshType                              = \"$mt\""
	
	set cxpath "DEM//c.DEM-MaterialTest//i.DEM-MeshPath"
	set mp [::xmlutils::setXml $cxpath "dv"]
	if {$mp eq ""} {set mp "0"}
	puts $fileid "MeshPath                              = \"$mp\""
	
	set cxpath "DEM//c.DEM-MaterialTest//i.DEM-ProbeLength"
	set pl [::xmlutils::setXml $cxpath "dv"]
	puts $fileid "SpecimenLength                           = $pl"
	
	set cxpath "DEM//c.DEM-MaterialTest//i.DEM-ProbeDiameter"
	set pd [::xmlutils::setXml $cxpath "dv"]
	puts $fileid "SpecimenDiameter                         = $pd"
    
    set ms [expr {3.14159265359*$pd*$pd/4}]
    puts $fileid "MeasuringSurface                         = $ms" 

}

proc ::wkcf::WritePostProcessData {fileid} {
    
    puts $fileid "# Results"
    puts $fileid ""
    
    # GraphExportFreq
    set cxpath "DEM//c.DEM-Results//c.DEM-Graphs//i.DEM-GraphExportFreq"
    set GEF [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "GraphExportFreq                   = $GEF"
    
    # Output Time Step
    set cxpath "DEM//c.DEM-Results//i.DEM-OTimeStep"
    set OTS [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "OutputTimeStep                   = $OTS"
	
    # PostDisplacement
    set cxpath "DEM//c.DEM-Results//c.DEM-PartElem//i.DEM-Displacement"
    set PDis [::xmlutils::setXml $cxpath "dv"]
    if {$PDis == "Yes"} {
	set PDis 1
    } else {
	set PDis 0
    }
    puts $fileid "PostDisplacement                 = \"$PDis\""
    
    # PostVelocity
    set cxpath "DEM//c.DEM-Results//c.DEM-PartElem//i.DEM-PostVel"
    set PVel [::xmlutils::setXml $cxpath "dv"]
    if {$PVel == "Yes"} {
	set PVel 1
    } else {
	set PVel 0
    }
    puts $fileid "PostVelocity                     = \"$PVel\""
    
     # PostTotalForces
    set cxpath "DEM//c.DEM-Results//c.DEM-PartElem//i.DEM-TotalForces"
    set PTF [::xmlutils::setXml $cxpath "dv"]
    if {$PTF == "Yes"} {
	set PTF 1
    } else {
	set PTF 0
    }
    puts $fileid "PostTotalForces                  = \"$PTF\""
    
    
     # PostTotalForces
    set cxpath "DEM//c.DEM-Results//c.DEM-PartElem//i.DEM-Rhs"
    set Rhs [::xmlutils::setXml $cxpath "dv"]
    if {$Rhs == "Yes"} {
	set Rhs 1
    } else {
	set Rhs 0
    }
    puts $fileid "PostRHS                  = \"$Rhs\""
    
    #        DampForces
    set cxpath "DEM//c.DEM-Results//c.DEM-PartElem//i.DEM-DampForces"
    set df [::xmlutils::setXml $cxpath "dv"]
    if {$df == "Yes"} {
	set df 1
    } else {
	set df 0
    }
    puts $fileid "PostDampForces                  = \"$df\""
    
    #        AppliedForces
    set cxpath "DEM//c.DEM-Results//c.DEM-PartElem//i.DEM-AppliedForces"
    set af [::xmlutils::setXml $cxpath "dv"]
    if {$af == "Yes"} {
	set af 1
    } else {
	set af 0
    }
    puts $fileid "PostAppliedForces                    = \"$af\""
    
    # Radius
    set cxpath "DEM//c.DEM-Results//c.DEM-PartElem//i.DEM-Radius"
    set rad [::xmlutils::setXml $cxpath "dv"]
    if {$rad == "Yes"} {
	set rad 1
    } else {
	set rad 0
    }
    puts $fileid "PostRadius                      = \"$rad\""
    
    # PostGroupId
    set cxpath "DEM//c.DEM-Results//c.DEM-PartElem//i.DEM-GroupId"
    set PGI [::xmlutils::setXml $cxpath "dv"]
    if {$PGI == "Yes"} {
	set PGI 1
    } else {
	set PGI 0
    }
    puts $fileid "PostGroupId                      = \"$PGI\""
    
    # PostExportId
    set cxpath "DEM//c.DEM-Results//c.DEM-PartElem//i.DEM-ExportId"
    set EID [::xmlutils::setXml $cxpath "dv"]
    if {$EID == "Yes"} {
	set EID 1
    } else {
	set EID 0
    }
    puts $fileid "PostExportId                       = \"$EID\""
    
    # PostExportSkinSphere
    set cxpath "DEM//c.DEM-Results//c.DEM-PartElem//i.DEM-ExportSkinSphere"
    set ESS [::xmlutils::setXml $cxpath "dv"]
    if {$ESS == "Yes"} {
	set ESS 1
    } else {
	set ESS 0
    }
    puts $fileid "PostExportSkinSphere                = \"$ESS\""
    
    # Calculate Rotations
    set cxpath "DEM//c.DEM-Options//c.DEM-AdvancedOptions//i.DEM-CalculateRotations"
    set useRotationOption [::xmlutils::setXml $cxpath "dv"]
    if {$useRotationOption == "Yes"} {
    
        # PostAngularVelocity 
        set cxpath "DEM//c.DEM-Results//c.DEM-PartElem//i.DEM-AngularVelocity"
        set av [::xmlutils::setXml $cxpath "dv"]
        if {$av == "Yes"} {
            set av 1
        } else {
            set av 0
        }
        puts $fileid "PostAngularVelocity                 = $av"   
    
        # PostParticleMoment 
        set cxpath "DEM//c.DEM-Results//c.DEM-PartElem//i.DEM-ParticleMoment"
        set pm [::xmlutils::setXml $cxpath "dv"]
        if {$pm == "Yes"} {
            set pm 1
        } else {
            set pm 0
        }
        puts $fileid "PostParticleMoment                 = $pm" 
    
        # PostEulerAngles 
       set cxpath "DEM//c.DEM-Results//c.DEM-PartElem//i.DEM-EulerAngles"
       set ea [::xmlutils::setXml $cxpath "dv"]
       if {$ea == "Yes"} {
	   set ea 1
       } else {
	   set ea 0
       }
       puts $fileid "PostEulerAngles                 = $ea"
    } else {
	puts $fileid "PostAngularVelocity                 = 0"
	puts $fileid "PostParticleMoment                 = 0"
	puts $fileid "PostEulerAngles                 = 0"
    }
    
    # Write all Dem Bond Elem Properties
    set basexpath "DEM//c.DEM-Results//c.DEM-BondElem"
    set ilist [::xmlutils::setXmlContainerIds $basexpath "Item"]
    # wa "ilist: $ilist"
    set kxpath "Applications/DEM"
    foreach item $ilist {
	set cxpath "DEM//c.DEM-Results//c.DEM-BondElem//i.$item"
	set value [::xmlutils::setXml $cxpath "dv"]
	# wa "$item: $value"
	if {$value == "Yes"} {
	    set value 1
	} else {
	    set value 0
	}
	set kid [string range $item 4 end]
	set kword [::xmlutils::getKKWord $kxpath $kid]
	puts $fileid "$kword                  = \"$value\""
    }
     
    puts $fileid ""
}

proc ::wkcf::WriteMaterialModelData {fileid} {
    set cxpath "DEM//c.DEM-ConstitutiveParameters//i.DEM-MaterialModel"
    set MaterialModel [::xmlutils::setXml $cxpath "dv"]

    puts $fileid "MaterialModel               = \"$MaterialModel\""
    
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-1DPlasticityAndDamage//i.DEM-LCS1"
    set val [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "LCS1                            = $val" 
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-1DPlasticityAndDamage//i.DEM-LCS2"
    set val [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "LCS2                            = $val" 
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-1DPlasticityAndDamage//i.DEM-LCS3"
    set val [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "LCS3                            = $val" 
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-1DPlasticityAndDamage//i.DEM-YRC1"
    set val [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "YRC1                            = $val" 
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-1DPlasticityAndDamage//i.DEM-YRC2"
    set val [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "YRC2                            = $val" 
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-1DPlasticityAndDamage//i.DEM-YRC3"
    set val [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "YRC3                            = $val"
    
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-1DPlasticityAndDamage//i.DEM-PlasticYoungModulus"
    set val [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "PlasticYoungModulus             = $val"
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-1DPlasticityAndDamage//i.DEM-Yields"
    set val [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "PlasticYieldStress              = $val"
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-1DPlasticityAndDamage//i.DEM-DamageDeformation"
    set val [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "DamageDeformationFactor         = $val"


    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-ExponentialHardening//i.Gamma1"
    set val [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "G1                              = $val" 
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-ExponentialHardening//i.Gamma2"
    set val [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "G2                              = $val" 
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-ExponentialHardening//i.Gamma3"
    set val [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "G3                              = $val" 
    set cxpath "DEM//c.DEM-ConstitutiveParameters//c.DEM-ExponentialHardening//i.MaxDeformation"
    set val [::xmlutils::setXml $cxpath "dv"]
    puts $fileid "MaxDef                          = $val" 
    
}

proc ::wkcf::WriteExplicitSolverVariables {} {
    # Write constitutive laws properties
    variable dprops;  variable ActiveAppList
    global KPriv
    
    set AppId "DEM"
    set filename "DEM_explicit_solver_var.py"
    set PDir [::KUtils::GetPaths "PDir"]
    
    set fullname [file native [file join $PDir $filename]]
    
    # First delete the file
    if {[file exists $fullname]} {
	set res [file delete -force -- $fullname]
    }

    if { [catch { set fileid [open $fullname w+] }] } {
	WarnWin [= "Cannot write file %s. Permission denied" $fullname].
	return 0
    }

    # Write the group properties
    set cproperty "dv"
    set rootid $AppId

    puts $fileid ""

    puts $fileid "# DEM General Options"
    puts $fileid ""
    
    # Dimension
    set cxpath "GeneralApplicationData//c.Domain//i.SpatialDimension"
    set ndim [string range [::xmlutils::setXml $cxpath $cproperty] 0 0]
    puts $fileid "Dimension                        = $ndim"
    
    
     # Get the drag correction due to sphericity
    set cxpath "$rootid//c.DEM-Options//c.Physical-opts//i.DragCorrection"
    set drag_modifier_type [::xmlutils::setXml $cxpath $cproperty]
    if {$drag_modifier_type eq "Chien"} {
	puts $fileid "drag_modifier_type               = 3"
    } else {
	puts $fileid "drag_modifier_type               = 2"
    }
    
    
     # Get if two way coupling is active
    set cxpath "$rootid//c.DEM-Fluid-interaction//i.DEM-TwoWayCoupling"
    set project_from_particles_option [::xmlutils::setXml $cxpath $cproperty]
    
    if {$project_from_particles_option eq "Yes"} {
       puts $fileid "project_from_particles_option        = 1"
    } else {
       puts $fileid "project_from_particles_option        = 0"
    }
    
     # Get if the lift force is active
    set cxpath "$rootid//c.DEM-Fluid-interaction//i.DEM-ConsiderLiftForce"
    set consider_lift_force_option [::xmlutils::setXml $cxpath $cproperty]
    
    if {$consider_lift_force_option eq "Yes"} {
       puts $fileid "consider_lift_force_option        = 1"
    } else {
       puts $fileid "consider_lift_force_option        = 0"
    }

    
    # Get the use bounding box 
    set cxpath "$rootid//c.DEM-Options//c.DEM-Boundingbox//i.UseBoundingBox"
    set UseBoundingBox [::xmlutils::setXml $cxpath $cproperty]
    if {$UseBoundingBox eq "Active"} {
	puts $fileid "BoundingBoxOption                = \"ON\""

	# Get the enlargement factor
	set cxpath "$rootid//c.DEM-Options//c.DEM-Boundingbox//i.EnlargementFactor"
	set EnlargementFactor [::xmlutils::setXml $cxpath $cproperty]
	puts $fileid "BoundingBoxEnlargementFactor     = $EnlargementFactor"
	
	# Get the bounding box type
	set cxpath "$rootid//c.DEM-Options//c.DEM-Boundingbox//i.BoundingBoxType"
	set BoundingBoxType [::xmlutils::setXml $cxpath $cproperty]
	if {$BoundingBoxType eq "Automatic"} {
	    puts $fileid "AutomaticBoundingBoxOption       = \"ON\""

	    # Write default bounding box values
	    ::wkcf::WriteBoundingBoxDefaults $fileid
	   
	} elseif {$BoundingBoxType eq "Fixed"} {
	    
	    puts $fileid "AutomaticBoundingBoxOption       = \"OFF\""
	    puts $fileid "BoundingBoxEnlargementFactor     = 1.0"
	    
	    # Get the bounding limit
	    set varlist [list MaxX MaxY MaxZ MinX MinY MinZ]
	    foreach varid $varlist {
		set cxpath "$rootid//c.DEM-Options//c.DEM-Boundingbox//i.$varid"
		set $varid [::xmlutils::setXml $cxpath $cproperty]
		puts $fileid "BoundingBox$varid                  = [set $varid]"
	    }
	}
    } else {
	puts $fileid "BoundingBoxOption                = \"OFF\""
	puts $fileid "BoundingBoxEnlargementFactor     = 1.0"
	puts $fileid "AutomaticBoundingBoxOption       = \"OFF\""
	
	# Write default bounding box values
	::wkcf::WriteBoundingBoxDefaults $fileid
    }
    puts $fileid ""
    
    # Get the DEM gravity
    set cxpath "$rootid//c.DEM-Options//c.DEM-Physical-opts//i.GravityValue"
    set usergravitymod [::xmlutils::setXml $cxpath $cproperty] 
    set cxpath "$rootid//c.DEM-Options//c.DEM-Physical-opts//i.Cx"
    set gravityCx [::xmlutils::setXml $cxpath $cproperty] 
    set cxpath "$rootid//c.DEM-Options//c.DEM-Physical-opts//i.Cy"
    set gravityCy [::xmlutils::setXml $cxpath $cproperty] 
    set cxpath "$rootid//c.DEM-Options//c.DEM-Physical-opts//i.Cz"
    set gravityCz [::xmlutils::setXml $cxpath $cproperty] 
	   
    set gravityvector [list $gravityCx $gravityCy $gravityCz] 
    set gravitymod [::MathUtils::VectorModulus $gravityvector]
    
    if {$gravitymod !=0.0} {
	   set endgravityCx [expr {($gravityCx*$usergravitymod)/$gravitymod}]
	   set endgravityCy [expr {($gravityCy*$usergravitymod)/$gravitymod}]
	   set endgravityCz [expr {($gravityCz*$usergravitymod)/$gravitymod}]
    } else {
	   set endgravityCx $gravityCx
	   set endgravityCy $gravityCy
	   set endgravityCz $gravityCz
    }
    puts $fileid "GravityX                         = $endgravityCx"
    puts $fileid "GravityY                         = $endgravityCy"
    puts $fileid "GravityZ                         = $endgravityCz"
    puts $fileid ""
    
    # Get the use velocity trap 
    set cxpath "$rootid//c.DEM-Results//c.DEM-VelocityTrap//i.UseVelocityTrap"
    set UseVelocityTrap [::xmlutils::setXml $cxpath $cproperty]
    if {$UseVelocityTrap eq "Active"} {
	puts $fileid "VelocityTrapOption                = \"ON\""

	# Get the bounding limit
	set varlist [list MaxX MaxY MaxZ MinX MinY MinZ]
	foreach varid $varlist {
		set cxpath "$rootid//c.DEM-Results//c.DEM-VelocityTrap//i.$varid"
		set $varid [::xmlutils::setXml $cxpath $cproperty]
		puts $fileid "VelocityTrap$varid                  = [set $varid]"
	    }        
	    
    } else {
	puts $fileid "VelocityTrapOption                = \"OFF\""
    }


    # Calculate Rotations
    set cxpath "$rootid//c.DEM-Options//c.DEM-AdvancedOptions//i.DEM-CalculateRotations"
    set useRotationOption [::xmlutils::setXml $cxpath $cproperty]
    if {$useRotationOption == "Yes"} {
	puts $fileid "RotationOption                   = \"ON\""
    } else {
	puts $fileid "RotationOption                   = \"OFF\""
    }
    
    puts $fileid ""
    
    # Dempack
    
    set cxpath "DEM//c.DEM-ConstitutiveParameters//i.DEM-MaterialModel"
    set MaterialModel [::xmlutils::setXml $cxpath $cproperty]
    
    if {$MaterialModel == "1DPlasticity"} {
    puts $fileid "Dempack                          = \"ON\""
    
    } else {
	 
	puts $fileid    "Dempack                          = \"OFF\""
    }

    
    # Clean IndentationsOption
    set cxpath "$rootid//c.DEM-Options//c.DEM-AdvancedOptions//i.DEM-CleanInitialIndentations"
    set CleanIndentationsOption [::xmlutils::setXml $cxpath $cproperty]
    if {$CleanIndentationsOption == "Yes"} {
        puts $fileid "CleanIndentationsOption          = \"ON\""
    
    } else {
        puts $fileid "CleanIndentationsOption          = \"OFF\""
    }
    puts $fileid ""
    
    
     # RemoveBallsInEmbedded
    set cxpath "$rootid//c.DEM-Options//c.DEM-AdvancedOptions//i.DEM-RemoveBallsInEmbedded"
    set RemoveBallsInEmbedded [::xmlutils::setXml $cxpath $cproperty]
    if {$RemoveBallsInEmbedded == "Yes"} {
        puts $fileid "RemoveBallsInEmbeddedOption          = 1"
    
    } else {
        puts $fileid "RemoveBallsInEmbeddedOption          = 0"
    }
    puts $fileid ""
    
    
    # Tangency Tolerance
    set cxpath "$rootid//c.DEM-Options//c.DEM-AdvancedOptions//i.DEM-TangencyTolerance"
    set useTangencyTolerance [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "DeltaOption                      = \"$useTangencyTolerance\""

    set cxpath "$rootid//c.DEM-Options//c.DEM-AdvancedOptions//i.DEM-TangencyToleranceValueAbsolute"
    set TangenTolerValue [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "SearchTolerance                  = $TangenTolerValue"

    set cxpath "$rootid//c.DEM-Options//c.DEM-AdvancedOptions//i.DEM-TangencyCoordinationNumber"
    set CoordNumber [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "CoordinationNumber        = $CoordNumber"
    
    
    # Amplified Search Radius Extension
    puts $fileid "AmplifiedSearchRadiusExtension   = 1.10000e+00"
    
    # Export Model Data Info
    set cxpath "$rootid//c.DEM-Options//c.DEM-AdvancedOptions//i.DEM-ExportModelDataInformation"
    set modatainfo [::xmlutils::setXml $cxpath $cproperty]
    if {$modatainfo eq "Active"} {
	puts $fileid "ModelDataInfo                    = \"ON\""
    } else {
	puts $fileid "ModelDataInfo                    = \"OFF\""
    }
    
    # Virtual Mass Coefficient
    set cxpath "$rootid//c.DEM-Options//c.DEM-AdvancedOptions//i.DEM-VirtualMassCoef"
    set vm [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "VirtualMassCoefficient           = $vm"
    
    # Rolling Friction
    set cxpath "$rootid//c.DEM-Options//c.DEM-AdvancedOptions//i.DEM-RollingFriction"
    set rf [::xmlutils::setXml $cxpath $cproperty]
    if {$rf eq "Activate"} {
	puts $fileid "RollingFrictionOption                    = \"ON\""
    } else {
	puts $fileid "RollingFrictionOption                    = \"OFF\""
    }

    
    # Dont search until failure
    set cxpath "$rootid//c.DEM-Options//c.DEM-AdvancedOptions//i.DEM-DontSearchUntilFail"
    set DS [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "DontSearchUntilFailure           = \"$DS\""

    # Contact Mesh Option
    set cxpath "$rootid//c.DEM-Results//i.DEM-ShowBondElements"
    set CMO [::xmlutils::setXml $cxpath $cproperty]
    if {$CMO eq "Yes"} {
    puts $fileid "ContactMeshOption                    = \"ON\""
    } else {
    puts $fileid "ContactMeshOption                    = \"OFF\""
    }    
    
    # Output file Type
    set cxpath "$rootid//c.DEM-Results//c.GiDOptions//i.GiDPostMode"
    set GiDMode [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "OutputFileType                   = \"$GiDMode\""
    
    # Multifile
    set cxpath "$rootid//c.DEM-Results//c.GiDOptions//i.GiDMultiFileFlag"
    set GiDMultiFileFlag [::xmlutils::setXml $cxpath $cproperty]
    # wa "multiflag $GiDMultiFileFlag"
    if {$GiDMultiFileFlag eq "Single"} {
        puts $fileid "Multifile                        = \"single_file\""
    } else {
        puts $fileid "Multifile                        = \"multiple_files\""
    }
    
    # Horizontal Fix Vel
    puts $fileid "HorizontalFixVel                 = \"ON\""
    
    ######################################################################################3
    
    puts $fileid " "
    puts $fileid "# Solution Strategy"
    
    # Integration Scheme
    set cxpath "$rootid//c.DEM-SolutionStrategy//i.DEM-TimeIntegrationScheme"
    set IntegScheme [::xmlutils::setXml $cxpath $cproperty]
    
    puts $fileid "IntegrationScheme                = \"$IntegScheme\""
    
    # Auto Reduction of Time Step Option
    set cxpath "$rootid//c.DEM-SolutionStrategy//c.DEM-TimeParameters//i.UseAutomaticDeltaTime"
    set DTimeSelection [::xmlutils::setXml $cxpath $cproperty]
    
    if {$DTimeSelection eq "Fixed"} {
	puts $fileid "AutoReductionOfTimeStepOption          = \"OFF\""
    } else {
	puts $fileid "AutoReductionOfTimeStepOption          = \"ON\""
    }
    set cxpath "$rootid//c.DEM-SolutionStrategy//c.DEM-TimeParameters//i.DeltaTimeSafetyFactor"
    set SafetyFact [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "DeltaTimeSafetyFactor                      = $SafetyFact"
    
    set cxpath "$rootid//c.DEM-SolutionStrategy//c.DEM-TimeParameters//i.DeltaTime"
    set MaxTimeStep [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "MaxTimeStep                      = $MaxTimeStep"
	  
    # Final Time
    set cxpath "$rootid//c.DEM-SolutionStrategy//c.DEM-TimeParameters//i.DEM-TotalTime"
    set TTime [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "FinalTime                        = $TTime"
    
    # Control Time
    set cxpath "$rootid//c.DEM-SolutionStrategy//c.DEM-TimeParameters//i.DEM-ScreenInfoOutput"
    set CTime [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "ControlTime                      = $CTime"
    
    # Neighbour Search Frequency
    set cxpath "$rootid//c.DEM-SolutionStrategy//c.DEM-TimeParameters//i.DEM-NeighbourSearchFrequency"
    set FrecTime [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "NeighbourSearchFrequency         = $FrecTime"
    
    ######################################################################################3
    
    puts $fileid "# Constitutive Parameters"
    
    # Material Model
    ::wkcf::WriteMaterialModelData   $fileid
    
    # Failiure Criterion
    puts $fileid "FailureCriterionType            = \"Uncoupled\""
    puts $fileid "AreaFactor                       = 1.00000e+00"
    set cxpath "$rootid//c.DEM-ConstitutiveParameters//c.DEM-FailureCriteria//i.DEM-TangentialStrength"
    set TangStr [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "TangentialStrength               = $TangStr"
    set cxpath "$rootid//c.DEM-ConstitutiveParameters//c.DEM-FailureCriteria//i.DEM-NormalTensileStrength"
    set NormalTens [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "NormalTensileStrength            = $NormalTens"
    set cxpath "$rootid//c.DEM-ConstitutiveParameters//c.DEM-FailureCriteria//i.DEM-InternalFrictionAngle"
    set IntFric [::xmlutils::setXml $cxpath $cproperty]
    puts $fileid "InternalFriction                 = $IntFric"
    
    # Damping
    ::wkcf::WriteDampingData $fileid
    
    puts $fileid "# Material Test"
    ::wkcf::WriteMatTestData $fileid

    puts $fileid ""
    puts $fileid "ElementType                      = \"[::wkcf::GetDEMActiveElements]\""
    
    puts $fileid "# PostProcess"
    ::wkcf::WritePostProcessData $fileid
    
    puts $fileid "# FROM CND:"
    puts $fileid ""
    puts $fileid "PredefinedSkinOption             = \"OFF\""
    
    set cxpath "$rootid//c.DEM-Elements//c.DEM-Element"
    set group [::xmlutils::setXmlContainerIds $cxpath]
    
    set cxpath "$rootid//c.DEM-Elements//c.DEM-Element//c.[lindex $group 0]//c.Properties//i.Material"
    set mat [::xmlutils::setXml $cxpath "dv"]
   
    if {$mat != ""} {
        set cxpath "DEMMaterial//m.$mat//p.Density"
        set densi [::xmlutils::setXml $cxpath "dv" "read" "" "mat"]
        
        set cxpath "DEMMaterial//m.$mat//p.YoungModulus"
        set young [::xmlutils::setXml $cxpath "dv" "read" "" "mat"]
        
        set cxpath "DEMMaterial//m.$mat//p.PoissonRatio"
        set poison [::xmlutils::setXml $cxpath "dv" "read" "" "mat"]
        
        set cxpath "DEMMaterial//m.$mat//p.ParticleFrictionAngle"
        set fricang [::xmlutils::setXml $cxpath "dv" "read" "" "mat"]
        
        puts $fileid " "
        puts $fileid "# WEATHERFORD"
        puts $fileid "w_densi                          = $densi"
        puts $fileid "w_young                          = $young"
        puts $fileid "w_poiss                          = $poison"
        puts $fileid "w_dynfrc                         = $fricang"
    }
    
    puts $fileid "MeanRadius                         = 0.0001"
    
    
    puts $fileid "# Declare Python Variables"
    puts $fileid ""

    set PName [::KUtils::GetPaths "PName"]
    puts $fileid "problem_name=\"${PName}\"" 
    # puts $fileid "problem_path=\"$PDir\"" 

    # Get the kratos path 
    set cxpath "GeneralApplicationData//c.ProjectConfiguration//i.KratosPath"
    set cproperty "dv"
    set KratosPath [::xmlutils::setXml $cxpath $cproperty]
    set KratosPath [file native $KratosPath]
    
    # Write the kratos path
    puts $fileid "kratos_path=\"${KratosPath}\""

    close $fileid
}

proc ::wkcf::WriteDemNodalVariables { AppId } {
    variable filechannel
    variable dprops
    variable ndime

    set nodelistbuffer [dict create ]
    
    # Write RADIUS
    set nodvar [::xmlutils::getKKWord "Applications/$AppId" "Radius"]
   
    set basexpath "$AppId//c.DEM-Elements"
    
    set gelemlist [::wkcf::GetDEMActiveElements]
    set nodelistbuffer ""
    
    foreach celemid $gelemlist {
	
	set elembasexpath "$basexpath//c.DEM-Element"
	# Get the group node list
	set grouplist [::xmlutils::setXmlContainerIds $elembasexpath]
	if {$grouplist != ""} {
	    foreach cgroupid $grouplist {
	      
		if {$cgroupid != ""} {
		    if {[GiD_EntitiesGroups get $cgroupid elements -count]} {
		        
		        GiD_File fprintf $filechannel "%s" "Begin NodalData $nodvar  // GUI group identifier: $cgroupid Elementid $celemid"
		        # De cada elemento necesitamos su nodo y su radio
		        foreach elem [GiD_EntitiesGroups get $cgroupid elements] {
		            set r 0
		            set info [GiD_Info list_entities -more elements $elem]
		            if { ![regexp {Type=([^ ]+) Nnode=([0-9]+) Volume=([^ ]+)} $info dummy type nnode volume] } {
		                set type ?
		                set nnode 0
		                set volume 0
		            }
		            if { $nnode > 0} {
		                set nodeid [lindex [GiD_Mesh get element $elem] 3]
		                dict lappend nodelistbuffer $cgroupid $nodeid
		                if {$ndime eq "3D"} {
		                    set r [expr {pow(3.0/4.0*$volume/$::MathUtils::Pi,1.0/3.0)}]
		                } else {
		                    set r [expr {sqrt($volume/$::MathUtils::Pi)}]
		                }
		                set r [GiD_FormatReal "%10.5e" $r]
		                GiD_File fprintf $filechannel "%s" "$nodeid 0 $r"
		            }
		        }
		        
		        GiD_File fprintf $filechannel "%s" "End NodalData"
		        GiD_File fprintf $filechannel "%s" ""
		    }
		    
		}
	    }
	}
    }
    
    # Write Cohesive Groups
    set nodvar [::xmlutils::getKKWord "Applications/$AppId" "Cohesive_Group"]
    
    set basexpath "$AppId//c.DEM-Cohesivegroup"
    set grouplist [::xmlutils::setXmlContainerIds $basexpath]
    
    foreach cgroupid $grouplist {
	if {$cgroupid != ""} {
	    if {[GiD_EntitiesGroups get $cgroupid elements -count]} {
		GiD_File fprintf $filechannel "%s" "Begin NodalData $nodvar  // GUI group identifier: $cgroupid"
		set elembasexpath "$basexpath//c.$cgroupid//c.Properties//i.CohGroup"
		# Get the group cohesive group
		set cohgroup [::xmlutils::setXml $elembasexpath "dv" ]
		
		if {[dict exists $nodelistbuffer $cgroupid]} {
		    foreach nodeid [dict get $nodelistbuffer $cgroupid] {
		        GiD_File fprintf $filechannel "%s" "$nodeid 0 $cohgroup"
		    }
		} else {
		    foreach elem [GiD_EntitiesGroups get $cgroupid elements] {
		        # Get the node from the element
		        set nodeid [lindex [GiD_Mesh get element $elem] 3]
		        dict lappend nodelistbuffer $cgroupid $nodeid
		        GiD_File fprintf $filechannel "%s" "$nodeid 0 $cohgroup"
		    }
		}
		GiD_File fprintf $filechannel "%s" "End NodalData"
		GiD_File fprintf $filechannel "%s" ""
	    }
	}
    }
    
    # Write All Material Dependent Properties; pe. Density, young modulus...
    set proplist [list "Density" "YoungModulus" "PoissonRatio" "ParticleFrictionAngle" "RestitutionCoefficient" "Color"]
    foreach cproperty $proplist {
	set nodvar [::xmlutils::getKKWord "Applications/$AppId" $cproperty]
	
	set basexpath "$AppId//c.DEM-Elements"
	
	foreach celemid $gelemlist {
	    
	    set elembasexpath "$basexpath//c.DEM-Element"
	    # Get the group node list
	    set grouplist [::xmlutils::setXmlContainerIds $elembasexpath]
	    if {$grouplist != ""} {
		foreach cgroupid $grouplist {
		    if {$cgroupid != ""} {
		        if {[GiD_EntitiesGroups get $cgroupid elements -count]} {
		            set matxpath "$elembasexpath//c.$cgroupid//c.Properties//i.Material"
		            set material [::xmlutils::setXml $matxpath "dv" ]
		            
		            set cxpath "DEMMaterial//m.$material//p.$cproperty"
		            set propvalue [::xmlutils::setXml $cxpath "dv" "read" "" "mat"]
		            
		            if {$cproperty eq "RestitutionCoefficient"} {
		                if {$propvalue == 0.0} {
		                    set propvalue 1.0
		                } else {
		                    set propvalue [expr {log($propvalue)}]
		                }
		            }
		            set propvalue [GiD_FormatReal "%10.5e" $propvalue]
		            GiD_File fprintf $filechannel "%s" "Begin NodalData $nodvar  // GUI group identifier: $cgroupid"
		            # De cada elemento necesitamos su nodo y su radio
		            if {[dict exists $nodelistbuffer $cgroupid]} {
		                foreach nodeid [dict get $nodelistbuffer $cgroupid] {
		                    GiD_File fprintf $filechannel "%s" "$nodeid 0 $propvalue"
		                }
		            } else {
		                foreach elem [GiD_EntitiesGroups get $cgroupid elements] {
		                    # Get the node from the element
		                    set nodeid [lindex [GiD_Mesh get element $elem] 3]
		                    dict lappend nodelistbuffer $cgroupid $nodeid
		                    GiD_File fprintf $filechannel "%s" "$nodeid 0 $propvalue"
		                }
		            }
		            
		            GiD_File fprintf $filechannel "%s" "End NodalData"
		            GiD_File fprintf $filechannel "%s" ""
		        }
		    }
		}
	    }
	}
    }
}

proc ::wkcf::WriteDemNodalVariables2 { AppId filechannel } {
    # Write RADIUS
    set nodvar [::xmlutils::getKKWord Applications/$AppId Radius]
    set basexpath $AppId//c.DEM-Elements
    set gelemlist [::wkcf::GetDEMActiveElements]
    set elembasexpath $basexpath//c.DEM-Element
    
    lassign [lindex [GiD_Info Mesh elements Sphere 1 end -array] 0] type element_ids element_nodes element_materials element_radii
    foreach element_id $element_ids element_node [lindex $element_nodes 0] element_radius $element_radii {
	set element_data($element_id) [list $element_node $element_radius]
    }
    #release memory
    foreach variable_name {element_ids element_nodes element_materials element_radii} {
	unset $variable_name
    }
    # Get the group node list
    set grouplist [::xmlutils::setXmlContainerIds $elembasexpath]
    if { [llength $grouplist] } {
	foreach celemid $gelemlist {
	    foreach cgroupid $grouplist {
		if { [GiD_EntitiesGroups get $cgroupid elements -count] } {
		    GiD_File fprintf $filechannel "Begin NodalData $nodvar  // GUI group identifier: $cgroupid Elementid $celemid"
		    foreach element_id [GiD_EntitiesGroups get $cgroupid elements] {
			lassign $element_data($element_id) element_node element_radius
			lappend group_nodes($cgroupid) $element_node
			GiD_File fprintf $filechannel "$element_node 0 $element_radius"
		    }
		    GiD_File fprintf $filechannel "End NodalData"
		    GiD_File fprintf $filechannel ""
		}
	    }
	}
    }
    # Write Cohesive Groups
    set nodvar [::xmlutils::getKKWord Applications/$AppId Cohesive_Group]
    set basexpath $AppId//c.DEM-Cohesivegroup
    set grouplist [::xmlutils::setXmlContainerIds $basexpath]
    foreach cgroupid $grouplist {
	if { [GiD_EntitiesGroups get $cgroupid elements -count] } {
	    GiD_File fprintf $filechannel "Begin NodalData $nodvar  // GUI group identifier: $cgroupid"
	    set elembasexpath $basexpath//c.$cgroupid//c.Properties//i.CohGroup
	    set cohesive_group [::xmlutils::setXml $elembasexpath dv]
	    if { [info exists group_nodes($cgroupid)] } {
		foreach node_id $group_nodes($cgroupid) {
		    GiD_File fprintf $filechannel "$node_id 0 $cohesive_group"
		}
	    } else {
		# Get the node from the element
		foreach element_id [GiD_EntitiesGroups get $cgroupid elements] {
		    set node_id [lindex $element_data($element_id) 0]
		    lappend group_nodes($cgroupid) $node_id
		    GiD_File fprintf $filechannel "$node_id 0 $cohesive_group"
		}
	    }
	    GiD_File fprintf $filechannel "End NodalData"
	    GiD_File fprintf $filechannel ""
	}
    }
    
    # Write All Material Dependent Properties; pe. Density, young modulus...
    set basexpath $AppId//c.DEM-Elements
    set elembasexpath $basexpath//c.DEM-Element
    set grouplist [::xmlutils::setXmlContainerIds $elembasexpath]
    foreach cproperty {Density YoungModulus PoissonRatio ParticleFrictionAngle RestitutionCoefficient Color} {
	set nodvar [::xmlutils::getKKWord Applications/$AppId $cproperty]
	foreach celemid $gelemlist {
	    foreach cgroupid $grouplist {
		if {[GiD_EntitiesGroups get $cgroupid elements -count]} {
		    set matxpath $elembasexpath//c.$cgroupid//c.Properties//i.Material
		    set material [::xmlutils::setXml $matxpath dv]
		    set cxpath DEMMaterial//m.$material//p.$cproperty
		    set propvalue [::xmlutils::setXml $cxpath dv read "" mat]
		    if {$cproperty eq "RestitutionCoefficient"} {
			if {$propvalue == 0.0} {
			    set propvalue 1.0
			} else {
			    set propvalue [expr {log($propvalue)}]
			}
		    }  
		    GiD_File fprintf $filechannel "Begin NodalData $nodvar  // GUI group identifier: $cgroupid"        
		    if { [info exists group_nodes($cgroupid)] } {
			foreach node_id $group_nodes($cgroupid) {
			    GiD_File fprintf $filechannel "$node_id 0 $propvalue"
			}
		    } else {
			# Get the node from the element
			foreach element_id [GiD_EntitiesGroups get $cgroupid elements] {
			    set node_id [lindex $element_data($element_id) 0]
			    lappend group_nodes($cgroupid) $node_id
			    GiD_File fprintf $filechannel "$node_id 0 $propvalue"
			}
		    }
		    GiD_File fprintf $filechannel "End NodalData"
		    GiD_File fprintf $filechannel ""
		}
	    }
	}
    }
    return 0
}


proc ::wkcf::WriteDEMConditionProperties { AppId cgroupid cpropvalue} {
    variable dprops; variable pflag
    variable filechannel
    
    # wa "Dem Properties AppId: $AppId cond: $cond cgroupid:$cgroupid"
    
    # For debug
    if {!$pflag} {
        set inittime [clock seconds]
    }
    
    # Kratos key word xpath
    set kxpath "Applications/$AppId"
    
    set proplist [list "VelocityBC-Ax" "VelocityBC-Vx"   \
		  "VelocityBC-Ay" "VelocityBC-Vy"   \
		  "VelocityBC-Az" "VelocityBC-Vz" "ForceIntegrationGroup" \
		  "TopGroup" "BottomGroup"]

    GiD_File fprintf $filechannel "%s" "  Begin MeshData // GUI Groupid: $cgroupid"
    GiD_File fprintf $filechannel ""
    foreach cprop $proplist cval $cpropvalue {
	set ckword [::xmlutils::getKKWord $kxpath $cprop]
	GiD_File fprintf $filechannel "%s" "    $ckword  $cval"
    }

    GiD_File fprintf $filechannel "%s" "  End MeshData"
    GiD_File fprintf $filechannel ""

     
    # For debug
    if {!$::wkcf::pflag} {
	set endtime [clock seconds]
	set ttime [expr {$endtime-$inittime}]
	# WarnWinText "endtime:$endtime ttime:$ttime"
	WarnWinText "Obtaining DEM-Inlet group using the new mesh format: [::KUtils::Duration $ttime]"
    }   
}


proc ::wkcf::WriteDEMFEMWallMeshProperties { AppId } {
    variable demfemchannel
    
    set rootid $AppId
    
    set basexpath "$rootid//c.DEM-Results//c.DEM-Graphs//c.DEM-ForceIntegrationGroup"
    set FIGgrlist [::xmlutils::setXmlContainerIds $basexpath]
    set basexpath "$AppId//c.DEM-MaterialTest//c.DEM-TopLayerGroup" 
    set TOPgrlist [::xmlutils::setXmlContainerIds $basexpath]
    set basexpath "$AppId//c.DEM-MaterialTest//c.DEM-BottomLayerGroup" 
    set BOTgrlist [::xmlutils::setXmlContainerIds $basexpath]
    
    set basexpath "$rootid//c.DEM-Conditions//c.DEM-FEM-Wall"
    set gproplist [::xmlutils::setXmlContainerIds $basexpath]
    
    foreach cgroupid $gproplist {
	# Get the group node list
	
	set nlist [::wkcf::GetDemFemWallGroupNodes $AppId $cgroupid]
	
	if {[llength $nlist]} {
	    set cproperty "dv"
	    set cxpath "${basexpath}//c.[list ${cgroupid}]//c.LinearVelocity//i.LinearVelocityX"
	    set LinearVelocityX [::xmlutils::setXml $cxpath $cproperty]   
	    set cxpath "${basexpath}//c.[list ${cgroupid}]//c.LinearVelocity//i.LinearVelocityY"
	    set LinearVelocityY [::xmlutils::setXml $cxpath $cproperty]   
	    set cxpath "${basexpath}//c.[list ${cgroupid}]//c.LinearVelocity//i.LinearVelocityZ"
	    set LinearVelocityZ [::xmlutils::setXml $cxpath $cproperty]  
	    set cxpath "${basexpath}//c.[list ${cgroupid}]//c.LinearVelocity//i.Periodic"
	    set IsPeriodic [::xmlutils::setXml $cxpath $cproperty] 
	    if {$IsPeriodic=="Yes"} {
		set cxpath "${basexpath}//c.[list ${cgroupid}]//c.LinearVelocity//i.Period"
		set Period [::xmlutils::setXml $cxpath $cproperty] 
	    } else {
		set Period "0.0"
	    }
	    set cxpath "${basexpath}//c.[list ${cgroupid}]//c.AngularVelocity//i.AngularVelocityX"
	    set AngularVelocityX [::xmlutils::setXml $cxpath $cproperty]   
	    set cxpath "${basexpath}//c.[list ${cgroupid}]//c.AngularVelocity//i.AngularVelocityY"
	    set AngularVelocityY [::xmlutils::setXml $cxpath $cproperty]   
	    set cxpath "${basexpath}//c.[list ${cgroupid}]//c.AngularVelocity//i.AngularVelocityZ"
	    set AngularVelocityZ [::xmlutils::setXml $cxpath $cproperty] 
	    set cxpath "${basexpath}//c.[list ${cgroupid}]//c.AngularVelocity//i.CenterOfRotationX"
	    set CenterOfRotationX [::xmlutils::setXml $cxpath $cproperty]   
	    set cxpath "${basexpath}//c.[list ${cgroupid}]//c.AngularVelocity//i.CenterOfRotationY"
	    set CenterOfRotationY [::xmlutils::setXml $cxpath $cproperty]   
	    set cxpath "${basexpath}//c.[list ${cgroupid}]//c.AngularVelocity//i.CenterOfRotationZ"
	    set CenterOfRotationZ [::xmlutils::setXml $cxpath $cproperty] 
	    set cxpath "${basexpath}//c.[list ${cgroupid}]//c.AngularVelocity//i.Periodic"
	    set AngularIsPeriodic [::xmlutils::setXml $cxpath $cproperty]
	    if {$AngularIsPeriodic=="Yes"} {
		set cxpath "${basexpath}//c.[list ${cgroupid}]//c.AngularVelocity//i.Period"
		set AngularPeriod [::xmlutils::setXml $cxpath $cproperty]
	    } else {
		set AngularPeriod "0.0"
	    }
	    
	    variable dem_fem_walls_property_number
	    set dem_fem_walls_property_number [expr $dem_fem_walls_property_number + 1 ]
	
	    GiD_File fprintf $demfemchannel "Begin Mesh $dem_fem_walls_property_number // GUI DEM-FEM-Wall group identifier: $cgroupid"  
	    
	    GiD_File fprintf $demfemchannel "Begin MeshData // GUI DEM-FEM-Wall group identifier: $cgroupid" 
	    GiD_File fprintf $demfemchannel "  VELOCITY \[3\] ($LinearVelocityX,$LinearVelocityY,$LinearVelocityZ)"
	    GiD_File fprintf $demfemchannel "  VELOCITY_PERIOD $Period"
	    GiD_File fprintf $demfemchannel "  ANGULAR_VELOCITY \[3\] ($AngularVelocityX,$AngularVelocityY,$AngularVelocityZ)"
	    GiD_File fprintf $demfemchannel "  ROTATION_CENTER \[3\] ($CenterOfRotationX,$CenterOfRotationY,$CenterOfRotationZ)"
	    GiD_File fprintf $demfemchannel "  ANGULAR_VELOCITY_PERIOD $AngularPeriod"
	    
	    set TOP 0
	    if {$cgroupid in $TOPgrlist} {set TOP 1}
	    GiD_File fprintf $demfemchannel "   [::xmlutils::getKKWord "Applications/$AppId" "TopGroup"]  $TOP"
	    set BOT 0
	    if {$cgroupid in $BOTgrlist} {set BOT 1}
	    GiD_File fprintf $demfemchannel "   [::xmlutils::getKKWord "Applications/$AppId" "BottomGroup"]  $BOT"
	    set FIG 0
	    if {$cgroupid in $FIGgrlist} {set FIG 1}
	    if {$BOT} {set FIG 1}
	    if {$TOP} {set FIG 1}
	    GiD_File fprintf $demfemchannel "   [::xmlutils::getKKWord "Applications/$AppId" "ForceIntegrationGroup"]  $FIG"

	    GiD_File fprintf $demfemchannel "End MeshData"
	    GiD_File fprintf $demfemchannel ""
	
	    GiD_File fprintf $demfemchannel "Begin MeshNodes"
	    foreach nodeid $nlist {                                                  
		GiD_File fprintf $demfemchannel "$nodeid"
	    }
	    GiD_File fprintf $demfemchannel "End MeshNodes"                                                           
	    GiD_File fprintf $demfemchannel "End Mesh"
	    GiD_File fprintf $demfemchannel ""
	}
    }
}


proc ::wkcf::WriteInletGroupMeshProperties { AppId } {
    # ABSTRACT: Write inlet condition group properties mdpa file (only the nodes)
    variable dprops
    variable deminletchannel
    
    # For debug
    if {!$::wkcf::pflag} {
	set inittime [clock seconds]
    }
    
    # Set the rootid
    set rootid "$AppId"
    
    # Get the values
    set basexpath "$rootid//c.DEM-Conditions//c.DEM-Inlet"
    set gproplist [::xmlutils::setXmlContainerIds $basexpath]
    # wa "gproplist:$gproplist"
    foreach cgroupid $gproplist {
	# Get the group node list
	set nlist [::wkcf::GetInletGroupNodes $AppId $cgroupid]
	if {[llength $nlist]} {
	    incr meshgroupid 1
	    # Get the meshid-group identifier mapping
	    set meshgroupid $dprops($AppId,Mesh,$cgroupid,MeshIdGroup) 
	    
	    # Write mesh properties for this group
	    GiD_File fprintf $deminletchannel "%s" "Begin Mesh $meshgroupid \/\/ GUI group identifier: $cgroupid"
	    # Write nodes
	    GiD_File fprintf $deminletchannel "%s" " "
	    GiD_File fprintf $deminletchannel "%s" " Begin MeshNodes"
	    foreach node_id $nlist {
		GiD_File fprintf $deminletchannel "%10i" $node_id
	    }
	    GiD_File fprintf $deminletchannel "%s" " End MeshNodes"
	    GiD_File fprintf $deminletchannel "%s" " "
	    GiD_File fprintf $deminletchannel "%s" "End Mesh"
	    GiD_File fprintf $deminletchannel "%s" ""
	}
    }
    
    # For debug
    if {!$::wkcf::pflag} {
	set endtime [clock seconds]
	set ttime [expr $endtime-$inittime]
	# WarnWinText "endtime:$endtime ttime:$ttime"
	WarnWinText "Write DEM-Inlet group using the new mesh format: [::KUtils::Duration $ttime]"
    }   
}

proc ::wkcf::GetDEMFEMElementName {firstelementid} {
    variable ndime
    
    set elemname "RigidFace"
    set elemname [append elemname $ndime]
    set elemname [append elemname [lindex [GiD_Mesh get element $firstelementid] 2] ]
    set elemname [append elemname "N"]
    
    return $elemname
    
    
}